/********************************************************************
 * chimeman.c A clock/timer software for DOS, Open Watcom version
 * Version: 1.0 (2007-08-07), Liang Zhao <writebear at gmail.com>
 * License: GNU GPL v2 or above, as you prefer.
 * Note: a VGA compatible card is required.
 * Compile: wcl chimaman.c (Open Watcom C/C++ required)
 * A Turbo C 2.01 version is also available with some nicer font.
 * But since that font is part of the Turbo C and may not be free,
 * we cannot distribute it on the Internet or by other medias.
 * This version is totally FREE (GPLed) and requires no dependence.
********************************************************************/

#include <conio.h>
#include <dos.h>
#include <graph.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

/* maximum length of strings */
#define MAX_STR_LEN 128

/* maximum number of sessions for mytimer */
#define MAX_SESSION 6

/* WIDTH of characters */
#define WIDTH 70

/* HEIGHT of characters */
#define HEIGHT 140

/**
 * global variables (constants, see graph.h for detail)
 */
short myColor[] = {7, 14, 12}; /* white, yellow, light red (red is 4) */
char * weekday[] = {"Sun", "Mon", "Tue", "Wed", "Thr", "Fri", "Sat"};
enum keyCode {KEY_ESC = 0x1B, KEY_SPACE = 0x20, KEY_ENTER = 0x0D, KEY_H = 0x68, KEY_Q = 0x71};

/**
 * global variables, not constant. Used for efficiency.
 * see function mytimer() for programming detail.
 */
int num; /* number of sessions for a timer instance */
char prompt[MAX_SESSION+1][MAX_STR_LEN]; /* prompt for each session */
long sec[MAX_SESSION]; /* time limit in second for each session */
int cnt[MAX_SESSION];  /* how many times should we chime? */
int freq[MAX_SESSION]; /* frequency (i.e., sound) of each chime */

/**
 * initialize the graphics card with EGA mode (2 pages)
 */
void initialize(void) {
	short errorcode;
#ifdef DEBUG
	FILE *fp = fopen("log", "w");
	struct videoconfig vc;
#endif
	_setvideomode(_ERESCOLOR); /* 640x350, 4 or 16 colors, 2 pages */
	errorcode = _grstatus();
	if (errorcode != _GROK) {
		printf("Error happened in setting the graphpics mode.\n");
		printf("[error code = %d]\n", errorcode);
		exit(-1);
	}
#ifdef DEBUG
	_getvideoconfig(&vc);
	fprintf(fp, "numxpixels = %d, numypixels = %d, numtextcols = %d, numtextrows = %d\n"
		"numcolors = %d, bitsperpixel = %d, numvideopages = %d\n"
		"mode = %d, adapter = %d, monitor = %d, memory = %d\n",
		vc.numxpixels, vc.numypixels, vc.numtextcols, vc.numtextrows,
		vc.numcolors, vc.bitsperpixel, vc.numvideopages,
		vc.mode, vc.adapter, vc.monitor, vc.memory);
#endif
}

/**
 * finalize = restore the old mode.
 */
void finalize() {
	_setvideomode(_DEFAULTMODE);
}

/**
 * function myclock(): a clock program
 * @param chime: should we chime (nonzero) at the start of every hour?
 */
void myclock(int chime) {
	int i, key;
	char msg[MAX_STR_LEN];
	struct tm *now;
	time_t t;
	int showHelp = 1;

	initialize();
	_setvisualpage(0);  /* Yes, we must use double buffers */
	for (i=0; ; i++) { /* overfow of i is no problem */
		if (kbhit()) {
			key = getch();
			if (key == KEY_ESC || key == KEY_Q) { /* exit */
				break;
			}
			else if (key == KEY_SPACE || key == KEY_ENTER) { /* turn on/off chime */
				chime = (chime + 1) % 2;
			}
			else if (key == KEY_H) { /* turn on/off help */
				showHelp = showHelp ? 0 : 1;
			}
			else { /* bad key touch */
			}
		}
		t = time(NULL);
		now = localtime(&t);
		_setactivepage((i+1)%2); /* draw on another video page */
		_clearscreen(_GCLEARSCREEN);
		if (showHelp) {
			_outtext("Key binding: ESC/q = quit, SPACE/ENTER = turn on/off chime, h = show/hide help");
		}
		sprintf(msg, "%04d-%02d-%02d %s%s\n", (*now).tm_year+1900, (*now).tm_mon+1, (*now).tm_mday, weekday[(*now).tm_wday], chime?" #":"");
		_setcharsize(HEIGHT/2-2, WIDTH/2);
		_grtext(50, 40, msg);
		sprintf(msg, "%02d:%02d:%02d", (*now).tm_hour, (*now).tm_min, (*now).tm_sec);
		_setcharsize(HEIGHT, WIDTH);
		_grtext(50, 160, msg);
		_setvisualpage((i+1)%2); /* show the page to users */
		if (chime && (*now).tm_min == 0 && (*now).tm_sec == 0) {
			sound(2000);
			sleep(1);
			nosound();
		}
		else {
			sleep(1);
		}
	}
	finalize();
}

/**
 * function mytimer(): a timer program
 * Global variables are used: num, prompt, sec, freq, ant cnt
 */
void mytimer(void) {
	char msg[MAX_STR_LEN];
	int i, n, key;
	struct tm *now;
	time_t finish, t;
	time_t paused; /* 0 for paused, otherwise the time when paused */
	int s, m, h;
	int showHelp = 1;

	initialize();
	finish = time(NULL);
	for (n=0; n<=num; n++) { /* n = num for the overtime session */
		if (n == 0) paused = time(NULL); /* pause at the start */
		if (n < num) { /* ie., not the overtime session */
			if (sec[n] < 0L) continue; /* pass invalid time */
			finish = time(NULL) + sec[n]; /* finish time */
		}
		_setvisualpage(0);
		for (i=0; ;i++) { /* overflow of i is no problem */
			if (kbhit()) {
				key = getch();
				if (key == KEY_ESC || key == KEY_Q) { /* exit */
					n = num + 1;
					break;
				}
				else if (key == KEY_SPACE) { /* pause */
					paused = paused ? 0 : time(NULL);
				}
				else if (key == KEY_ENTER && paused) { /* reset */
					n = -1; /* after next break, n is 0 */
					break;
				}
				else if (key == KEY_H) { /* turn on/off help */
					showHelp = showHelp ? 0 : 1;
				}
				else {
				}
			}
			t = time(NULL);
			if (paused) { /* extend the finish time */
				finish += t - paused;
				paused = t;
			}
			if (n<num && t>finish) break;
			now = localtime(&t);
			_setactivepage((i+1)%2);
			_clearscreen(_GCLEARSCREEN);
			_setcolor((n<num)?myColor[n%sizeof(myColor)]:4);
			if (showHelp) {
				_outtext("Key binding: ESC/q = quit, SPACE = pause, ENTER = reset when paused, h = help");
			}
			_setcharsize(HEIGHT/3, WIDTH/3);
			if (strlen(prompt[n]) == 0) {
				sprintf(msg, "%04d-%02d-%02d %s %02d:%02d:%02d\n", (*now).tm_year+1900, (*now).tm_mon+1, (*now).tm_mday, weekday[(*now).tm_wday], (*now).tm_hour, (*now).tm_min, (*now).tm_sec);
				_grtext(50, 30, msg);
			}
			else {
				_grtext(50, 30, prompt[n]);
			}
			t = (n<num) ? (finish-t) : (t-finish);
			s = (int)(t % 60); /* second */
			t = (t - s) / 60;
			m = (int)(t % 60); /* minute */
			t = (t - m) / 60;
			h = (int)(t % 100); /* hour, round into two digits */
			sprintf(msg, "%02d:%02d:%02d", h, m, s);
			_setcharsize(HEIGHT, WIDTH);
			_grtext(60, 160, msg);
			_setvisualpage((i+1)%2);
			sleep(1);
		}
		if (n == num - 1) finish = time(NULL);
		if (n >= 0 && n < num) {
			for (i=0; i<cnt[n]; i++) {
				sound(freq[n]);
				sleep(1);
				nosound();
				sleep(1);
			}
		}
	}
	finalize();
}

/**
 * help function: just print a few help message
 * @param name name of the program
 */
void help(char *name) {
	printf("ChimeMan 1.0 for DOS (OpenWatcom) (c) 2007 Liang Zhao <writebear at gmail.com>\n"
	  "Usage: %s [options]\n"
	  "  No option means a clock with chime (to turn off/on, use SPACE after start).\n"
	  "  Otherwise it performs as a timer.\n"
	  "  Option 'sec' is a timer of sec seconds with default settings.\n"
	  "  Option 'prompt sec freq cnt' is for a sec-second timer with prompt \n"
	  "      and sound of frequency freq repeated cnt times.\n"
	  "  Note: up to %d groups of 'prompt sec freq cnt' are OK.\n",
	  name, MAX_SESSION
	);
}

/**
 * main function
 * @param argc number of argument
 * @param argv command line parameters
 * see help() for usage.
 */
int main(int argc, char * argv[]) {
	switch (argc) {
		case 1:  /* default is a clock with chime at each hour */
			myclock(1);
			break;
		case 2:  /* otherwise it is a simple timer */
			sec[0] = atol(argv[1]);
			if (sec[0] <= 0) {
				help(argv[0]);
				return 1;
			}
			num = 1;
			strncpy(prompt[0], "", 2); /* default prompt */
			strncpy(prompt[1], "", 2); /* must be set */
			cnt[0] = 3;                /* repeat three times */
			freq[0] = 2000;            /* frequency 2000hz */
			mytimer();
			break;
		default: /* Or a complicatd timer */
			if (argc % 4 != 1) { /* argv is not enough */
				help(argv[0]);
				return 2;
			}
			for (num=0; num<argc/4 && num<MAX_SESSION; num++) {
				strncpy(prompt[num], argv[num*4+1], MAX_STR_LEN);
				sec[num] = atol(argv[num*4+2]);
				freq[num] = atoi(argv[num*4+3]);
				cnt[num] = atoi(argv[num*4+4]);
			}
			strncpy(prompt[num], "Overtime", MAX_STR_LEN);
			mytimer();
			break;
	}
	return 0;
}
