/********************************************************************
 * chimeman.c A clock/timer software for (Free)DOS. Turbo C version.
 * Version: 1.0 (2007-08-07), Liang Zhao <writebear at gmail.com>
 * License: GNU GPL v2 or above, at your will. Notice that this does
 *          not cover egavga.bgi and trip.chr which belongs to the
 *          Turbo C 2.01 -- now free of charge for PERSONAL use.
 *          Use the OpenWatcom version for a purely free package.
 *          There is no functional difference but the font is dirty.
 * Compile: tcc -N -w chimaman.c graphics.lib (Turbo C 2.01 required)
 * NOTICE: the sleep() function works fine but the delay() function
 *         is aweful. Never use delay() with Turbo C!
 * Usage: make sure the egavga.bgi and trip.chr are in the same
 *        directory of chimeman.exe. You can create a bootable CD
 *        or floppy or USB stick with FreeDOS and this package.
 * Note: a VGA compatible card is required to run this software.
********************************************************************/

#include <conio.h>
#include <dos.h>
#include <graphics.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

/* maximum length of strings */
#define MAX_STR_LEN 128

/* maximum number of sessions for mytimer */
#define MAX_SESSION 6

/* base scale (dividor) for resizing fonts */
#define SCALE 10

/* xscale to resize the width of characters */
/* 42 means 42/10 = 4.2 times bigger. */
#define XSCALE 42

/* yscale to resize the height of characters */
#define YSCALE 50

/**
 * global variables (constants, see graphics.h for detail)
 */
int graphdriver = VGA;
int graphmode = VGAMED; /* 640x350, 16 colors, 2 pages */
int font = TRIPLEX_FONT; /* other choices: SANS_SERIF_FONT, GOTHIC_FONT */
char * weekday[] = {"Sun", "Mon", "Tue", "Wed", "Thr", "Fri", "Sat"};
int myColor[] = {WHITE, YELLOW, LIGHTRED};
enum keyCode {KEY_ESC = 0x1B, KEY_SPACE = 0x20, KEY_ENTER = 0x0D, KEY_H = 0x68, KEY_Q = 0x71};

/**
 * global variables, not constant. Used for efficiency.
 * see function mytimer() for programming detail.
 */
int num; /* number of sessions for a timer instance */
char prompt[MAX_SESSION+1][MAX_STR_LEN]; /* prompt for each session */
long sec[MAX_SESSION]; /* time limit in second for each session */
int cnt[MAX_SESSION];  /* how many times should we chime? */
int freq[MAX_SESSION]; /* frequency (i.e., sound) of each chime */

/**
 * function prototypes (required by the strict ANSI C)
 */
void initialize(void); /* initialize the graphics card */
void finalize(void);   /* close the graphics library */
void myclock(int);     /* the clock function */
void mytimer(void);    /* the timer function */
void help(char *);     /* output usage message */

/**
 * initialize the graphics card
 */
void initialize(void) {
	int errorcode;

	initgraph(&graphdriver, &graphmode, "");
	errorcode = graphresult();
	if (errorcode != grOk) {
		printf("Error happened in setting the graphpics mode.\n");
		printf("[%s]\n", grapherrormsg(errorcode));
		exit(-1);
	}
}

/**
 * finalize = close the graphics libaray
 */
void finalize() {
	closegraph();
}

/**
 * function myclock(): a clock program
 * @param chime: should we chime (nonzero) at the start of every hour?
 */
void myclock(int chime) {
	int i, key;
	char msg[MAX_STR_LEN];
	struct tm *now;
	time_t t;
	int showHelp = 1;

	initialize();
	settextstyle(font, HORIZ_DIR, USER_CHAR_SIZE);
	setvisualpage(0);  /* Yes, we must use double buffers */
	for (i=0; ; i++) { /* overfow of i is no problem */
		if (kbhit()) {
			key = getch();
			if (key == KEY_ESC || key == KEY_Q) { /* exit */
				break;
			}
			else if (key == KEY_SPACE || key == KEY_ENTER) { /* turn on/off chime */
				chime = (chime + 1) % 2;
			}
			else if (key == KEY_H) {
				showHelp = !showHelp;
			}
			else  { /* bad key touch */
			}
		}
		t = time(NULL);
		now = localtime(&t);
		setactivepage((i+1)%2); /* draw on another video page */
		cleardevice();
		if (showHelp) {
			settextstyle(DEFAULT_FONT, HORIZ_DIR, 1);
			outtextxy(0, 0, "Key binding: ESC/q = quit, SPACE/ENTER = turon on/off chime, h = show/hide help");
			settextstyle(font, HORIZ_DIR, USER_CHAR_SIZE);
		}
		setusercharsize(XSCALE/2-2, SCALE, YSCALE/2-2, SCALE);
		sprintf(msg, "%04d-%02d-%02d %s%s", (*now).tm_year+1900, (*now).tm_mon+1, (*now).tm_mday, weekday[(*now).tm_wday], chime?" #":"");
		outtextxy(60, 30, msg);
		sprintf(msg, "%02d:%02d:%02d", (*now).tm_hour, (*now).tm_min, (*now).tm_sec);
		setusercharsize(XSCALE, SCALE, YSCALE, SCALE);
		outtextxy(90, 150, msg);
		setvisualpage((i+1)%2); /* show the page to users */
		if (chime && (*now).tm_min == 0 && (*now).tm_sec == 0) {
			sound(2000);
			sleep(1);
			nosound();
		}
		else {
			sleep(1);
		}
	}
	finalize();
}

/**
 * function mytimer(): a timer program
 * Next global variables are used: num, prompt, sec, freq, cnt.
 */
void mytimer(void) {
	char msg[MAX_STR_LEN];
	int i, n, key;
	struct tm *now;
	time_t finish, t;
	time_t paused; /* 0 if not paused, otherwise the time when paused. */
	int s, m, h;
	int showHelp = 1;

	initialize();
	settextstyle(font, HORIZ_DIR, USER_CHAR_SIZE);
	finish = time(NULL);
	for (n=0; n<=num; n++) { /* n = num for the overtime session */
		if (n == 0) paused = time(NULL); /* we pause at the start */
		if (n < num) { /* i.e., not the overtime session */
			if (sec[n] <= 0L) continue; /* pass invalid time */
			finish = time(NULL) + (time_t)sec[n]; /* finish time */
		}
		setvisualpage(0);
		for (i=0; ; i++) { /* overflow of i is not a problem */
			if (kbhit()) {
				key = getch();
				if (key == KEY_ESC || key == KEY_Q) { /* exit */
					n = num + 1;
					break;
				}
				else if (key == KEY_SPACE) { /* pause */
					paused = (paused) ? 0 : time(NULL);
				}
				else if (key == KEY_ENTER && paused) { /* reset */
					n = -1; /* after break, n will be 0 */
					break;
				}
				else if (key == KEY_H) {
					showHelp = !showHelp;
				}
				else { /* bad key input */
				}
			}
			t = time(NULL);
			if (paused) { /* extend the finish time if paused */
				finish += (t - paused);
				paused = t;
			}
			if (n<num && t>finish) break;
			now = localtime(&t);
			setactivepage((i+1)%2);
			cleardevice();
			setcolor((n<num)?myColor[num % sizeof(myColor)]:RED);
			if (showHelp) {
				settextstyle(DEFAULT_FONT, HORIZ_DIR, 1);
				outtextxy(0, 0, "Key binding: ESC/q = quit, SPACE = pause, Enter = reset when paused, h = help");
				settextstyle(font, HORIZ_DIR, USER_CHAR_SIZE);
			}
			setusercharsize(XSCALE/4, SCALE, YSCALE/4, SCALE);
			if (strlen(prompt[n]) == 0) {
				sprintf(msg, "%04d-%02d-%02d %s %02d:%02d:%02d", (*now).tm_year+1900, (*now).tm_mon+1, (*now).tm_mday, weekday[(*now).tm_wday], (*now).tm_hour, (*now).tm_min, (*now).tm_sec);
				outtextxy(80, 30, msg);
			}
			else {
				outtextxy(80, 30, prompt[n]);
			}
			t = (n<num) ? (finish - t) : (t - finish);
			s = (int)(t % 60); /* second */
			t = (t - s) / 60;
			m = (int)(t % 60); /* minutes */
			t = (t - m) / 60;
			h = (int)(t % 60); /* hour */
			h = h % 100; /* round it into two digits */
			sprintf(msg, "%02d:%02d:%02d", h, m, s);
			setusercharsize(XSCALE, SCALE, YSCALE, SCALE);
			outtextxy(80, 120, msg);
			setvisualpage((i+1)%2);
			sleep(1);
		}
		if (n == num -1) finish = time(NULL);
		if (n >= 0 && n < num) {
			for (i=0; i<cnt[n]; i++) {
				sound(freq[n]);
				sleep(1);
				nosound();
				sleep(1);
			}
		}
	}
	finalize();
}

/**
 * help function: just print a few help message
 * @param name name of the program
 */
void help(char *name) {
	printf("ChimeMan 1.0 for DOS (Turbo C) (c) 2007 Liang Zhao <writebear at gmail.com>\n"
	  "Usage: %s [options]\n"
	  "  No option means a clock with chime (to turn off/on, use SPACE after start).\n"
	  "  Otherwise it performs as a timer.\n"
	  "  Option 'sec' is a timer of sec seconds with default settings.\n"
	  "  Option 'prompt sec freq cnt' is for a sec-second timer with prompt  and \n"
	  "      sound of frequency freq repeated cnt times.\n"
	  "  Note: up to %d groups of 'prompt sec freq cnt' are OK.\n",
	  name, MAX_SESSION
	);
}

/**
 * main function
 * @param argc number of argument
 * @param argv command line parameters
 * see help() for usage.
 */
int main(int argc, char * argv[]) {
	switch (argc) {
		case 1:  /* default is a clock with chime at each hour */
			myclock(1);
			break;
		case 2:  /* otherwise it is a simple timer */
			sec[0] = atol(argv[1]);
			if (sec[0] <= 0) {
				help(argv[0]);
				return 1;
			}
			num = 1;
			strncpy(prompt[0], "", 2); /* default prompt */
			strncpy(prompt[1], "", 2); /* must be set */
			cnt[0] = 3;                /* repeat three times */
			freq[0] = 2000;            /* frequency 2000hz */
			mytimer();
			break;
		default: /* Or a complicatd timer */
			if (argc % 4 != 1) { /* argv is not enough */
				help(argv[0]);
				return 2;
			}
			for (num=0; num<argc/4 && num<MAX_SESSION; num++) {
				strncpy(prompt[num], argv[num*4+1], MAX_STR_LEN);
				sec[num] = atol(argv[num*4+2]);
				freq[num] = atoi(argv[num*4+3]);
				cnt[num] = atoi(argv[num*4+4]);
			}
			strncpy(prompt[num], "Overtime", MAX_STR_LEN);
			mytimer();
			break;
	}
	return 0;
}
